import random
import string
from django.conf import settings
from django.db import models
from django.dispatch import receiver
from django.db.models.signals import pre_save, post_save

from djangoldp.models import Model
from .permissions import ProjectPermissions, ProjectMemberPermissions


class Customer(Model):
    name = models.CharField(max_length=255)
    address = models.CharField(max_length=255, null=True, blank=True)
    logo = models.URLField(blank=True, null=True)
    companyRegister = models.CharField(max_length=255, null=True, blank=True)
    firstName = models.CharField(max_length=255, null=True, blank=True)
    lastName = models.CharField(max_length=255, null=True, blank=True)
    role = models.CharField(max_length=255, null=True, blank=True)
    email = models.EmailField(null=True, blank=True)
    phone = models.CharField(max_length=255, null=True, blank=True)

    def __str__(self):
        return self.name


class BusinessProvider(Model):
    name = models.CharField(max_length=255)
    fee = models.PositiveIntegerField(default='0')

    def __str__(self):
        return self.name


def auto_increment_project_number():
  last_inc = Project.objects.all().order_by('id').last()
  if not last_inc:
    return 1
  return last_inc.number + 1


class Project(Model):
    name = models.CharField(max_length=255)
    description = models.TextField(null=True, blank=True)
    number = models.PositiveIntegerField(default=auto_increment_project_number, editable=False)
    creationDate = models.DateField(auto_now_add=True)
    customer = models.ForeignKey(Customer, on_delete=models.DO_NOTHING)  # WARN add import
    team = models.ManyToManyField(settings.AUTH_USER_MODEL, through='Member', blank=True)
    captain = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.DO_NOTHING, blank=True, null=True,
                                related_name='+')
    driveID = models.TextField(null=True, blank=True)
    businessProvider = models.ForeignKey(BusinessProvider, blank=True, null=True, on_delete=models.DO_NOTHING)
    jabberID = models.CharField(max_length=255, blank=True, null=True)
    jabberRoom = models.BooleanField(default=True)

    class Meta:
        nested_fields = ['team', 'customer', 'members']
        authenticated_perms = ['view', 'add', 'change', 'delete']
        permission_classes = [ProjectPermissions]
        rdf_type = 'hd:project'

    def __str__(self):
        return self.name

    def get_admins(self):
        return self.members.filter(is_admin=True)


class Member(Model):
    project = models.ForeignKey(Project, on_delete=models.CASCADE, related_name='members')
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    name = models.CharField(max_length=255, null=True, blank=True)
    is_admin = models.BooleanField(default=False)

    class Meta:
        container_path = "project-members/"
        permission_classes = [ProjectMemberPermissions]
        authenticated_perms = ["view"]
        unique_together = ['user', 'project']

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        # cannot be duplicated Members
        if not self.pk and Member.objects.filter(project=self.project, user=self.user).exists():
            return

        super(Member, self).save(*args, **kwargs)


@receiver(pre_save, sender=Project)
def set_jabberid(sender, instance, **kwargs):
    if settings.JABBER_DEFAULT_HOST and not instance.jabberID:
        instance.jabberID = '{}@conference.{}'.format(
            ''.join(
                [
                    random.choice(string.ascii_letters + string.digits)
                    for n in range(12)
                ]
            ).lower(),
            settings.JABBER_DEFAULT_HOST
        )
        instance.jabberRoom = True


@receiver(post_save, sender=Project)
def set_captain_as_member(instance, created, **kwargs):
    # add captain as an admin member, if they've not already been added
    if created and instance.captain is not None and\
            not instance.members.filter(user=instance.captain).exists():
        Member.objects.create(user=instance.captain, project=instance, is_admin=True, name=instance.captain.name())



from djangoldp.permissions import LDPPermissions


# auxiliary function tests user is an admin for specified project
def is_user_admin_of_project(user, project):
    from .models import Member

    try:
        project_member = Member.objects.get(user=user, project=project)
        return project_member.is_admin

    except:
        return False


class ProjectPermissions(LDPPermissions):
    def has_permission(self, request, view):
        # anonymous users have no rights
        if request.user.is_anonymous and not request.method == 'OPTIONS':
            return False

        # request on an existing resource - this will be reviewed by has_object_permission
        if request.method == 'PATCH' or request.method == 'DELETE' or request.method == 'PUT':
            return True

        return super().has_permission(request, view)

    def has_object_permission(self, request, view, obj):
        from .models import Member

        # anonymous users have no rights
        if request.user.is_anonymous and not request.method == 'OPTIONS':
            return False

        # admins have full permissions
        if is_user_admin_of_project(request.user, obj):
            return True

        # other members can perform GET only
        if request.method != 'GET':
            return False

        if not Member.objects.filter(user=request.user, project=obj).exists():
            return False

        return super().has_object_permission(request, view, obj)


class ProjectMemberPermissions(LDPPermissions):
    def has_permission(self, request, view):
        from djangoldp.models import Model

        # anonymous users have no rights
        if request.user.is_anonymous and not request.method == 'OPTIONS':
            return False

        # request on an existing resource - this will be reviewed by has_object_permission
        if request.method == 'PATCH' or request.method == 'DELETE' or request.method == 'PUT':
            return True

        # only admins can add new members to a project
        if request.method == 'POST':
            obj = Model.resolve_id(request._request.path)
            return is_user_admin_of_project(request.user, obj.project)

        return super().has_permission(request, view)

    def has_object_permission(self, request, view, obj):
        # anonymous users have no rights
        if request.user.is_anonymous and not request.method == 'OPTIONS':
            return False

        # admins have full permissions
        if is_user_admin_of_project(request.user, obj.project):
            if request.method == 'DELETE':
                # I cannot remove myself if I am the last admin
                if obj.pk == request.user.pk:
                    if obj.project.get_admins().count() == 1:
                        return False

                # I cannot remove another admin
                elif obj.is_admin:
                    return False

            return True

        # I can remove myself
        if obj.user.pk == request.user.pk:
            return True

        return super().has_object_permission(request, view, obj)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.26 on 2019-12-09 17:12
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('djangoldp_project', '0002_member_is_admin'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='member',
            options={},
        ),
        migrations.AlterField(
            model_name='customer',
            name='address',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='customer',
            name='companyRegister',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='customer',
            name='email',
            field=models.EmailField(blank=True, max_length=254, null=True),
        ),
        migrations.AlterField(
            model_name='customer',
            name='firstName',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='customer',
            name='lastName',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='customer',
            name='phone',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='customer',
            name='role',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='member',
            name='name',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='project',
            name='description',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='project',
            name='driveID',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AlterUniqueTogether(
            name='member',
            unique_together=set([('user', 'project')]),
        ),
    ]



import factory
from .models import Customer, Project, Member, BusinessProvider
from django.db.models.signals import post_save

@factory.django.mute_signals(post_save)
class CustomerFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Customer

    name = factory.Faker('company')
    address = factory.Faker('address')
    logo = factory.Faker('url')
    companyRegister = factory.Faker('random_int', min=0)
    firstName = factory.Faker('first_name')
    lastName = factory.Faker('last_name')
    role = factory.Faker('job')
    email = factory.Faker('email')
    phone = factory.Faker('phone_number')

class BusinessProviderFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = BusinessProvider

    name = factory.Faker('name')
    fee = factory.Faker('random_int', min=0, max=10)

@factory.django.mute_signals(post_save)
class ProjectFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Project

    name = factory.Faker('word')
    description = factory.Faker('text', max_nb_chars=250)
    number = factory.Faker('random_int', min=0)
    creationDate = factory.Faker('past_date')
    customer = factory.SubFactory(CustomerFactory)
    businessProvider = factory.SubFactory(BusinessProviderFactory)
    jabberID = factory.Faker('email')
    jabberRoom = True

    @factory.post_generation
    def members(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for member in extracted:
                Member.objects.create(name=factory.Faker('job'), project=self, user=member)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-10-26 16:01
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import djangoldp.fields
import djangoldp_project.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='BusinessProvider',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('urlid', djangoldp.fields.LDPUrlField(blank=True, null=True, unique=True)),
                ('name', models.CharField(max_length=255)),
                ('fee', models.PositiveIntegerField(default='0')),
            ],
            options={
                'abstract': False,
                'default_permissions': ('add', 'change', 'delete', 'view', 'control'),
            },
        ),
        migrations.CreateModel(
            name='Customer',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('urlid', djangoldp.fields.LDPUrlField(blank=True, null=True, unique=True)),
                ('name', models.CharField(max_length=255)),
                ('address', models.CharField(blank=True, max_length=255)),
                ('logo', models.URLField(blank=True, null=True)),
                ('companyRegister', models.CharField(blank=True, max_length=255)),
                ('firstName', models.CharField(blank=True, max_length=255)),
                ('lastName', models.CharField(blank=True, max_length=255)),
                ('role', models.CharField(blank=True, max_length=255)),
                ('email', models.EmailField(blank=True, max_length=254)),
                ('phone', models.CharField(blank=True, max_length=255)),
            ],
            options={
                'abstract': False,
                'default_permissions': ('add', 'change', 'delete', 'view', 'control'),
            },
        ),
        migrations.CreateModel(
            name='Member',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('urlid', djangoldp.fields.LDPUrlField(blank=True, null=True, unique=True)),
                ('name', models.CharField(blank=True, max_length=255)),
            ],
            options={
                'abstract': False,
                'default_permissions': ('add', 'change', 'delete', 'view', 'control'),
            },
        ),
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('urlid', djangoldp.fields.LDPUrlField(blank=True, null=True, unique=True)),
                ('name', models.CharField(max_length=255)),
                ('description', models.TextField(blank=True)),
                ('number', models.PositiveIntegerField(default=djangoldp_project.models.auto_increment_project_number, editable=False)),
                ('creationDate', models.DateField(auto_now_add=True)),
                ('driveID', models.TextField(blank=True)),
                ('jabberID', models.CharField(blank=True, max_length=255, null=True)),
                ('jabberRoom', models.BooleanField(default=True)),
                ('businessProvider', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='djangoldp_project.BusinessProvider')),
                ('captain', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('customer', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, to='djangoldp_project.Customer')),
                ('team', models.ManyToManyField(blank=True, through='djangoldp_project.Member', to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.AddField(
            model_name='member',
            name='project',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='members', to='djangoldp_project.Project'),
        ),
        migrations.AddField(
            model_name='member',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2019-11-28 16:15
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djangoldp_project', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='member',
            name='is_admin',
            field=models.BooleanField(default=False),
        ),
    ]
